
import argparse
import sys
import os
import zipfile
from pathlib import Path
from narizaka.data import InputData


columns = 'audio,speaker_id,sentence,duration'
def run():
    parser = argparse.ArgumentParser(description = 'Utility to make audio dataset from  audio and text book')

    
    parser.add_argument('data', type=Path, help='This is path to data directory where each subdirectory\n'
                        'contains only one text file(book format or just text) and audio file/files for this text. '
                        'Audio files could have free folder structure.')
    parser.add_argument('-t',  required=False, type=Path, help='Path to text file(book format or just text)', default=None)
    parser.add_argument('-o',  type=Path, help='Output directory', default=Path('./output/'))
    parser.add_argument('-device',  type=str, help='Device to run on', default='auto')
    parser.add_argument('-c', action='store_true',  help='Cache only mode', default=False)
    parser.add_argument('-n',  type=int, help='Limit number of CPU workers', default=None)
    parser.add_argument('-sr',  type=int, help='Resample to', default=24000)
    parser.add_argument('-columns',  type=str, help=f'Columns to include, default values is "{columns}", this is all possible columns', default=columns)



    args = parser.parse_args()
    if not args.data.is_dir():
        print("-data argument should point to the directory")
        sys.exit(-1)

        
    input_data = InputData(args)
    if not input_data.is_empty():
        print(f"The following books have been found:")
        
        if len(input_data.transcribed_books):
            print('\nFully transcribed:')
            for book_pair in input_data.transcribed_books:
                print(book_pair.audio_book.speaker_id, book_pair.text_book_path)
        if len(input_data.needs_transcribe_books):
            print('\nPartially or not transcribed:')
            for book_pair in input_data.needs_transcribe_books:
                print(book_pair.audio_book.speaker_id, book_pair.text_book_path)

        input_data.process()

        if args.c:
            if not args.o.exists():
                os.makedirs(args.o, exist_ok=True)
            archive_path = args.o/(args.data.name +'.zip')
            with zipfile.ZipFile(archive_path , mode="w") as archive:
                for p in input_data.get_all_pairs():
                    for cache_file_path in p.audio_book.get_cache_files():
                        archive.write(cache_file_path, arcname=f'narizaka/'+ cache_file_path.name)
            print(f'\nCache archive have been saved to {archive_path}')
    else:
        print('Have not found any data to process.')



if __name__ == '__main__':
    run()




import magic
import sys
from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor, as_completed
from narizaka.audiobook import AudioBook
from narizaka.transcriber import Transcriber
from narizaka.aligner import Aligner
from narizaka.utils import AudioTextPair


class InputData():
    supported_mimes =[
            'application/epub+zip',
            'text/xml',
            'text/plain',
            'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
        ]

    def __init__(self, args):
        self.args = args
        self.transcribed_books = []
        self.needs_transcribe_books = []
        self.transcriber = None
        found_book = None
        if args.t:
            mimetype = magic.from_file(filename=args.t, mime=True)
            if mimetype in self.supported_mimes:
                found_book = args.t
            else:
                print('Text file is not suported.\n suported formats are:', self.supported_mimes)
                sys.exit(-1)
        else:
            for item in args.data.iterdir():
                if not item.is_dir():
                    mimetype = magic.from_file(filename=item, mime=True)
                    if mimetype in self.supported_mimes:
                        found_book = item
                        break
        if found_book:
            self._make_and_add_book_pair(AudioBook(args.data), item)
            return

        items = list(args.data.iterdir())
        for speaker_id, book_or_group in enumerate(tqdm(items, desc='Discovering data')):
            if book_or_group.is_dir():
                if found_book:=self._find_one_book(book_or_group):
                    self._make_and_add_book_pair(AudioBook(book_or_group,  speaker_id=speaker_id), found_book)
                else:
                    for group_item in book_or_group.iterdir():
                        if group_item.is_dir():
                            if found:=self._find_one_book(group_item):
                                self._make_and_add_book_pair(AudioBook(group_item,  speaker_id=speaker_id), found)

    def _find_one_book(self, book_dir):
        for book_item in book_dir.iterdir():
                if not book_item.is_dir():
                    mimetype = magic.from_file(filename=book_item, mime=True)
                    if mimetype in self.supported_mimes:
                        return book_item
        return None 
    
    def get_all_pairs(self):
        return self.transcribed_books + self.needs_transcribe_books

    def _make_and_add_book_pair(self, audio_book, text_book_path):
        p = AudioTextPair(audio_book, text_book_path)
        if audio_book.is_transcribed():
            self.transcribed_books.append(p)
        else:
            t = self._get_transcriber()
            t.add(p)
            self.needs_transcribe_books.append(p)

    def _get_transcriber(self):
        if not self.transcriber:
            self.transcriber = Transcriber(self.args)
        return self.transcriber
    
    def is_empty(self):
        return not len(self.transcribed_books) and not len(self.needs_transcribe_books)

    def process(self):
        futures = []
        total_result = [0,0]
        with ProcessPoolExecutor(max_workers=self.args.n) as pool:
            if not self.args.c:
                for pair in self.transcribed_books:
                    future = pool.submit(Aligner(self.args).run, pair)
                    futures.append(future)
            if self.needs_transcribe_books:
                transcriber = self._get_transcriber()
                for pair in tqdm(transcriber.transcribe(), total=len(self.needs_transcribe_books), desc='Transcribing'):
                    if not self.args.c:
                        future = pool.submit(Aligner(self.args).run, pair)
                        futures.append(future)
            if futures:
                progress = tqdm(total=len(futures), desc='Aligning and cropping')
                for future in as_completed(futures):
                    recognized, total, name = future.result()
                    total_result[0] += recognized
                    total_result[1] += total
                    progress.write(f'Result for book {name}:')
                    progress.write(f'Extracted {recognized/3600:.3f} hours from audio duration of {total/3600:.3f}')
                    progress.write(f'It is {(recognized/total)*100:.1f}% of total audio\n')
                    progress.update(1)
                progress.close()

                print('\nTotal statistic:')
                print(f'Extracted {total_result[0]/3600:.3f} hours from audio duration of {total_result[1]/3600:.3f}')
                print(f'It is {(total_result[0]/total_result[1])*100:.1f}% of total audio\n')



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from . import (lang_UK)

CONVERTER_CLASSES = {
    
    'uk': lang_UK.Num2Word_UK(),

}

CONVERTES_TYPES = ['cardinal', 'ordinal', 'ordinal_num', 'year', 'currency']


def num2words(number, ordinal=False, lang='en', to='cardinal', **kwargs):
    # We try the full language first
    if lang not in CONVERTER_CLASSES:
        # ... and then try only the first 2 letters
        lang = lang[:2]
    if lang not in CONVERTER_CLASSES:
        raise NotImplementedError()
    converter = CONVERTER_CLASSES[lang]

    if isinstance(number, str):
        number = converter.str_to_number(number)

    # backwards compatible
    if ordinal:
        to = 'ordinal'

    if to not in CONVERTES_TYPES:
        raise NotImplementedError()

    return getattr(converter, 'to_{}'.format(to))(number, **kwargs)



import os
import pathlib
import torch
import multiprocessing
from collections import OrderedDict
from dataclasses import asdict
from multiprocessing import Queue, Process
from queue import Empty
from narizaka.utils import convert_media
from narizaka.asr_backends.faster_whisper import FasterWhisperTranscriber

multiprocessing.set_start_method('spawn', force=True)

class BackendWorker(Process):
    def __init__(self, device, device_index, book_files_queue, output_queue) -> None:
        
        self.book_files_queue = book_files_queue
        self.transcribed = output_queue
        self.device = device
        self.device_index = device_index
        super().__init__()
    

    def run(self):
        self.backend = FasterWhisperTranscriber(self.device, self.device_index)
        try:
            while book_files := self.book_files_queue.get(timeout=10):
                result = self.backend.transcribe(book_files[1])
                self.transcribed.put_nowait({'text_book_filename': book_files[0], 
                                            'audio_file': book_files[1], 
                                            'transcription': result})
        except Empty:
            print('Done worker thread')




class Transcriber():
    def __init__(self, device) -> None:
        self.cache_path = pathlib.Path(os.getenv('HOME', '.')) / '.cache/narizaka'
        if not self.cache_path.exists():
            os.makedirs(self.cache_path, exist_ok=True)

        self.device = 'cuda' if torch.cuda.device_count() > 0 and device != 'cpu' else 'cpu'
        self.audio_files_q = Queue()
        self.transcribed = Queue()
        self.devices = torch.cuda.device_count() if torch.cuda.device_count() > 0 and device != 'cpu' else 2
        self.books = {}

    
    def add(self, pair):
        self.books[pair.text_book_path] = pair
        for audio_file in pair.audio_book.non_transcribed_files():
            self.audio_files_q.put_nowait((pair.text_book_path, audio_file['filename']))
    
    def transcribe(self):
        self.workers = []
        for index in range(self.devices):
            worker = BackendWorker(self.device, index if self.device == 'cuda' else 0, self.audio_files_q, self.transcribed)
            worker.start()
            self.workers.append(worker)

        while any([w.is_alive() for w in self.workers]):
            try:
                transcribed = self.transcribed.get(timeout=20)
                pair = self.books[transcribed['text_book_filename']]
                pair.audio_book.save_transcription(transcribed['audio_file'], transcribed['transcription'] )
                if pair.audio_book.is_transcribed():
                    yield pair
            except Empty:
                pass



import os
import ffmpeg
import tempfile
from retry import retry
from narizaka.audiobook import AudioBook


class AudioTextPair():
    def __init__(self, audio_book:AudioBook, text_book_path:str):
        self.audio_book = audio_book
        self.text_book_path = text_book_path
    

@retry(tries=3, delay=1)
def convert_media(filename, format='flac', sr=None):
    probe = ffmpeg.probe(filename)
    if probe.get('format').get('format_name') != format or probe['streams'][0]['sample_rate'] != str(sr):
        fl, audio_file = tempfile.mkstemp(suffix=f'.{format}')
        os.close(fl)
        stream = ffmpeg.input(filename)
        if sr:
            stream = ffmpeg.output(stream, audio_file, acodec='pcm_s16le' if format == 'wav' else 'flac' , ar=sr, ac=1, loglevel='panic')
        else:
            stream = ffmpeg.output(stream, audio_file,  acodec='pcm_s16le' if format == 'wav' else 'flac' , ac=1, loglevel='panic')
        ffmpeg.run(stream, overwrite_output=True)
        return audio_file, sr or int(probe['streams'][0]['sample_rate'])
    else:
        return filename, int(probe['streams'][0]['sample_rate'])


import regex
from .num2words import num2words
import unicodedata

simple_replacements = {
    '№' : 'номер',
    '§': 'номер'
}

masc_replacments_dict = {
    '%':['відсоток', 'відсотки', 'відсотків'],
    'мм': ['міліметр', 'міліметри', 'міліметрів'],
    'см': ['сантиметр', 'сантиметри', 'сантиметрів'],
    'мм': ['міліметр', 'міліметри', 'міліметрів'],
    # 'м': ['метр', 'метри', 'метрів'],
    'км': ['кілометр', 'кілометри', 'кілометрів'],
    'гц': ['герц', 'герци', 'герців'],
    'кгц': ['кілогерц', 'кілогерци', 'кілогерців'],
    'мгц': ['мегагерц', 'мегагерци', 'мегагерців'],
    'ггц': ['гігагерц', 'гігагерци', 'гігагерців'],
    'вт': ['ват', 'вати', 'ватів'],
    'квт': ['кіловат', 'кіловати', 'кіловатів'],
    'мвт': ['мегават', 'мегавати', 'мегаватів'],
    'гвт': ['гігават', 'гігавати', 'гігаватів'],
    'дж': ['джоуль', 'джоулі', 'джоулів'],
    'кдж': ['кілоджоуль', 'кілоджоулі', 'кілоджоулів'],
    'мдж': ['мегаджоуль', 'мегаджоулі', 'мегаджоулів'],
    'см2': ['сантиметр квадратний', 'сантиметри квадратні', 'сантиметрів квадратних'],
    'м2': ['метр квадратний', 'метри квадратні', 'метрів квадратних'],
    'м2': ['кілометр квадратний', 'кілометри квадратні', 'кілометрів квадратних'],
    '$': ['долар', 'долари', 'доларів'],
    '€': ['євро', 'євро', 'євро'],
}

fem_replacments_dict = {
    'кал': ['калорія', 'калорії', 'калорій'],
    'ккал': ['кілокалорія', 'кілокалорії', 'кілокалорій'],
    'грн': ['гривня', 'гривні', 'гривень'],
    'грв': ['гривня', 'гривні', 'гривень'],
    '₴': ['гривня', 'гривні', 'гривень'],
}

neu_replacments_dict = {
     '€': ['євро', 'євро', 'євро'],
}

all_replacments_keys = list(masc_replacments_dict.keys()) + list(fem_replacments_dict.keys()) + list(neu_replacments_dict.keys())

#Ordinal types
#Називний
ordinal_nominative_masculine_cases = ('й','ий')
ordinal_nominative_feminine_cases = ('a','ша', 'я')
ordinal_nominative_neuter_cases = ('е',)

#Родовий
ordinal_genitive_masculine_case = ('го','о',)
ordinal_genitive_feminine_case = ('ї', 'ої')


#Давальний
ordinal_dative_masculine_case = ('му',)
ordinal_dative_feminine_case = ('й','ій')

#Знахідний
ordinal_accusative_masculine_case = ordinal_genitive_masculine_case
ordinal_accusative_feminine_case = ('у',)

#Орудний
ordinal_instrumental_masculine_case = ('им', 'ім')
ordinal_instrumental_feminine_case = ('ю')


#Місцевий
# ordinal_locative_masculine_case = ordinal_dative_masculine_case
# ordinal_locative_feminine_case = ordinal_dative_feminine_case

numcases_r = regex.compile(rf'((?:^|\s)(\d+)\s*(\-?)(([^\d,]*?)|(\-\.+))(?:\.|,|:|-)?)(\s+[^,.:\-]|$)', regex.IGNORECASE, regex.UNICODE)

cardinal_genitive_endings = ('а', 'e', 'є', 'й')
ordinal_genitive_cases = ('року',)

def number_form(number):
    if number[-1] == "1":
        return 0
    elif number[-1] in ("2", "3", "4"):
        return 1
    else:
        return 2

def replace_cases(number, dash, case='', next_word=''):
    gender = 'masculine'
    m_case = 'nominative'
    to = 'ordinal'
    repl = ''
    if not dash:
        if case in all_replacments_keys:
            if case in masc_replacments_dict.keys():
                repl = masc_replacments_dict.get(case)[number_form(number)]
                gender = 'masculine'
            elif case in fem_replacments_dict.keys():
                repl = fem_replacments_dict.get(case)[number_form(number)]
                gender = 'feminine'
            elif case in neu_replacments_dict.keys():
                repl = neu_replacments_dict.get(case)[number_form(number)]
                gender = 'neuter'
            to = 'cardinal'
        else:
            if len(case) < 3 and case and case[-1] in cardinal_genitive_endings:
                m_case = 'genitive'
                gender='masculine'
                to = 'cardinal'
            elif case in ordinal_genitive_cases:
                to = 'ordinal'
                m_case = 'genitive'
                repl = case
            else:
                to = 'cardinal'
                repl = case
                
    else:    
        if case in ordinal_nominative_masculine_cases:
            m_case = 'nominative'
            gender = 'masculine'
        elif case in ordinal_nominative_feminine_cases:
            m_case = 'nominative'
            gender = 'feminine'
        elif case in ordinal_nominative_neuter_cases:
            m_case = 'nominative'
            gender = 'neuter'
        elif case in ordinal_genitive_masculine_case:
            m_case = 'genitive'
            gender = 'masculine'
        elif case in ordinal_genitive_feminine_case:
            m_case = 'genitive'
            gender = 'feminine'
        elif case in ordinal_dative_masculine_case:
            m_case = 'dative'
            gender = 'masculine'
        elif case in ordinal_dative_feminine_case:
            m_case = 'dative'
            gender = 'feminine'
        elif case in ordinal_accusative_feminine_case:
            m_case = 'accusative'
            gender = 'feminine'
        elif case in ordinal_instrumental_masculine_case:
            m_case = 'instrumental'
            gender = 'masculine'
        elif case in ordinal_instrumental_feminine_case:
            m_case = 'instrumental'
            gender = 'feminine'
        else:
            if case and case[-1] in cardinal_genitive_endings:
                m_case = 'genitive'
                gender='masculine'
                to = 'cardinal'
                repl = case
            else:
                #print(f'UNKNOWN CASE {number}-{case}')
                pass
    return_str = ''
    try:
        return_str = num2words(number, to=to, lang='uk', case=m_case, gender=gender)
    except:
        print(f'Exception in numebr conversion: {number}, {to}, {m_case}, {gender}')
    if repl:
        return_str +=  ' ' + repl
    if not next_word or (next_word and  next_word.strip().isupper()):
        return_str += '.'
    return return_str

def norm(text):
    text = regex.sub(r'[\t\n]', ' ', text)
    text = regex.sub(rf"[{''.join(simple_replacements.keys())}]", lambda x: f' {simple_replacements[x.group()]} ', text)
    text = regex.sub(r"(\d)\s+(\d)", r"\1\2", text)
    text = regex.sub(r'\s+', ' ', text)
    text = unicodedata.normalize('NFC', text)
    matches = numcases_r.finditer(text)
    pos = 0
    new_text = ''
    for m in matches:
        repl = replace_cases(m.group(2), m.group(3), m.group(4), m.group(7))
        new_text += text[pos:m.start(0)]+ ' ' + repl
        pos = m.end(1)
    new_text += text[pos:]
    return new_text.strip()

    

#1-го квітня, на 1-му поверсі Яринка загубила 2грн але знайшла 5€. Але її 4-річна сестричка забрала 50% її знахідки.
#Також 2003 року щось там сталося і 40-річний чоловік помер. Його знайшли через 3 години.

#01:51:37.250 -> 01:51:44.650:  Серед міленіалів цей показник становить 39%, серед покоління X – 30%,
#39
#30
#MATCHED: серед міленіалів цей показник становить тридцять девять , серед покоління Х - тридцять ,
#Skipped because contains inapropirate characters

#05:28:52.350 -> 05:29:00.000:  2016 рік завершився з чистими збитками 1,2 мільярди доларів США.
#2016
#MATCHED: дві тисячі шістнадцять рік завершився з чистими збитками 1,2 млрд доларів США.


import auditok

class Splitter():
    def split_to_segments(self, audio_file, all_words):
        def _split(region, threshold=46, deep=4):
            if not region.meta:
                region.meta = {'start': 0}
            audio_regions = []
            for r in region.split(
                min_dur=0.2,     # minimum duration of a valid audio event in seconds
                max_dur=80,       # maximum duration of an event
                max_silence=0.11, # maximum duration of tolerated continuous silence within an event
                energy_threshold=threshold # threshold of detection
            ):
                r.meta = {'start': r.meta.start+region.meta.start, 'end': r.meta.end+region.meta.start}
                if r.duration > 8.0 and deep:
                    regions = _split(r, threshold+2, deep-1)
                    if len(regions)> 1:
                        audio_regions = audio_regions + regions
                else:
                    audio_regions.append(r)
            return audio_regions

        
        region = auditok.load(str(audio_file), large_file=True)
        audio_regions = sorted(_split(region), key=lambda x: x.meta.start)
        
        pugaps = []
        text = ''
        for i, word in enumerate(all_words[:-1]):
            text += word['word']
            if word['word'][-1] in [',','.','?','-',':','!', '»', ';'] or \
              ((all_words[i+1]['start'] - word['end']) > 0.2 and (all_words[i+1]['end'] - all_words[i+1]['start']) > 0.4 and
              (word['end'] - word['start']) > 0.4):
                pugaps.append([word['end'], all_words[i+1]['start'], i])
                text = ''
        
        
        temp_reg = None
        start_word = 0
        regions_by_punct = []
        for i, r in enumerate(audio_regions[:-1]):

            if not temp_reg:
                temp_reg = r
            else:
                start = temp_reg.meta.start
                temp_reg += r
                temp_reg.meta = {'end': r.meta.end, 'start': start}

            gap_dur = audio_regions[i+1].meta.start - r.meta.end
            gap_point = r.meta.end + (gap_dur/2)
            found = next((item for item in pugaps if (item[0]-0.1) <= gap_point <= item[1]+0.1), None)
            
            if found:
                if start_word != found[2]+1:
                    text = ''.join([word['word'] for word in all_words[start_word:found[2]+1]])
                    if text.strip():
                        regions_by_punct.append({'start': temp_reg.meta.start,
                                                'end': r.meta.end,
                                                'text': text})
                        temp_reg = None
                        start_word = found[2]+1
                

        ready_segment = {}
        for segment in regions_by_punct:
            if not ready_segment:
                ready_segment = segment
                continue

            if ((segment['start'] - ready_segment['end']) < 0.4 and ready_segment['text'][-1] in [',', ':', '-', '»', '\'', '.', '?', '!'] and (segment['end'] - ready_segment['start']) < 11)\
                or (ready_segment['text'][-1] in [',', ':', '-', '»', '\'']  and (segment['end'] - ready_segment['start']) < 20):
                ready_segment['end'] = segment['end']
                ready_segment['text'] += segment['text']
            else:
                yield ready_segment
                ready_segment = segment
        yield ready_segment

    


import os
import magic
import ffmpeg
import pathlib
import hashlib
import json
from collections import OrderedDict
from glob import glob



class AudioBook():
    def __init__(self, filename: pathlib.Path, speaker_id=None) -> None:
        self.cache_path = pathlib.Path(os.getenv('HOME', '.')) / '.cache/narizaka'
        self.speaker_id = speaker_id
        self.files = OrderedDict()
        self.duration = 0.0
        if not filename.exists():
            raise Exception('Audio path doesn\'t exists')

        if filename.is_dir():
            files = list(glob(f'{filename.resolve()}/**', recursive=True))
            files.sort()
            for f in files:
                f = pathlib.Path(f)
                if not f.is_dir() and self._is_media(f):
                    self._add_file(f)
            if not self.files:
                raise Exception('Directory doesn\'t contain any audio files')
        else:
            if not self._is_media(filename):
                raise Exception('Audio book is not audio file')    
            self._add_file(filename)


    def _add_file(self, filename):
        if self._is_media(filename):
            hash, cache_file = self._calc_hash(filename)
            probe = ffmpeg.probe(filename)
            self.duration += float(probe["format"]["duration"])
            file_dict = {'filename': filename, 'transcribed': False}
            if cache_file.exists():
                file_dict['transcribed'] = True
            self.files[hash] = file_dict
    
    def _calc_hash(self, filename):
        with open(filename, 'rb') as f:
            data = f.read()
            hash = hashlib.sha256(data).hexdigest()
            return hash, self.cache_path / pathlib.Path(hash+'.json')
    
    def is_transcribed(self):
        return not list(self.non_transcribed_files())
    
    def non_transcribed_files(self):
        return filter(lambda x: x['transcribed'] == False, self.files.values())

    def save_transcription(self, filename, transcription):
        hash, cache_file = self._calc_hash(filename)
        if not hash in self.files:
            raise Exception('Can\'t find file in untranscribed list')
        with open(cache_file, 'w') as w:
            data = []
            for r in transcription.all_words():
                data.append(r.to_dict())
            w.write(json.dumps(data, ensure_ascii=False, indent=4))
        self.files[hash]['transcribed'] = True

    def _is_media(self, filename: pathlib.Path)-> bool:
        m = magic.detect_from_filename(filename=filename)
        return  True if (m.mime_type.split('/')[0] == 'audio' or m.mime_type.split('/')[0] == 'video') else True if 'audio' in m.name.lower() else False

    def transcription(self):
        for hash, val in self.files.items():
            cache_file = self.cache_path / pathlib.Path(hash+'.json')
            if not val['transcribed']:
                raise Exception(f'{val["filename"]} is not transcribed')
            yield val['filename'], json.load(open(cache_file, 'r'))

    def get_cache_files(self):
        for hash in self.files.keys():
            yield self.cache_path / pathlib.Path(hash+'.json')


from narizaka.languages.uk.textnormalizer import norm

def test_uah():
    assert norm('просто 1 грн.') == 'просто одна гривня.'
    assert norm('1 грн. тест') == 'одна гривня тест'
    assert norm('Просто 1 грн. тест') == 'Просто одна гривня тест'
    assert norm('просто 1 грн.') == 'просто одна гривня.'
    assert norm('1 грн. тест') == 'одна гривня тест'
    assert norm('Просто 1 грн. тест') == 'Просто одна гривня тест'

    assert norm('просто 2 грн.') == 'просто дві гривні.'
    assert norm('2 грн. тест') == 'дві гривні тест'
    assert norm('Просто 2 грн. тест') == 'Просто дві гривні тест'
    assert norm('просто 2 грн.') == 'просто дві гривні.'
    assert norm('2 грн. тест') == 'дві гривні тест'
    assert norm('Просто 2 грн. тест') == 'Просто дві гривні тест'

    assert norm('просто 6 грн.') == 'просто шість гривень.'
    assert norm('6 грн. тест') == 'шість гривень тест'
    assert norm('Просто 6 грн. тест') == 'Просто шість гривень тест'
    assert norm('просто 6 грн.') == 'просто шість гривень.'
    assert norm('6 грн. тест') == 'шість гривень тест'
    assert norm('Просто 6 грн. тест') == 'Просто шість гривень тест'

    assert norm('просто 6 547 грн.') == 'просто шість тисяч п\'ятсот сорок сім гривень.'
    assert norm('6 547 грн. тест') == 'шість тисяч п\'ятсот сорок сім гривень тест'
    assert norm('Просто 6 547 грн. тест') == 'Просто шість тисяч п\'ятсот сорок сім гривень тест'
    assert norm('просто 6 547 грн.') == 'просто шість тисяч п\'ятсот сорок сім гривень.'
    assert norm('6 547 грн. тест') == 'шість тисяч п\'ятсот сорок сім гривень тест'
    assert norm('Просто 6 547 грн. тест') == 'Просто шість тисяч п\'ятсот сорок сім гривень тест'
    
    assert norm('просто 1 грн') == 'просто одна гривня.'
    assert norm('1 грн тест') == 'одна гривня тест'
    assert norm('Просто 1 грн тест') == 'Просто одна гривня тест'
    assert norm('просто 1 грн') == 'просто одна гривня.'
    assert norm('1 грн тест') == 'одна гривня тест'
    assert norm('Просто 1 грн тест') == 'Просто одна гривня тест'

    assert norm('просто 2 грн') == 'просто дві гривні.'
    assert norm('2 грн тест') == 'дві гривні тест'
    assert norm('Просто 2 грн тест') == 'Просто дві гривні тест'
    assert norm('просто 2 грн') == 'просто дві гривні.'
    assert norm('2 грн тест') == 'дві гривні тест'
    assert norm('Просто 2 грн тест') == 'Просто дві гривні тест'

    assert norm('просто 6 грн') == 'просто шість гривень.'
    assert norm('6 грн тест') == 'шість гривень тест'
    assert norm('Просто 6 грн. тест') == 'Просто шість гривень тест'
    assert norm('просто 6 грн') == 'просто шість гривень.'
    assert norm('6 грн тест') == 'шість гривень тест'
    assert norm('Просто 6 грн тест') == 'Просто шість гривень тест'

    assert norm('просто 6 547 грн') == 'просто шість тисяч п\'ятсот сорок сім гривень.'
    assert norm('6 547 грн тест') == 'шість тисяч п\'ятсот сорок сім гривень тест'
    assert norm('Просто 6 547 грн тест') == 'Просто шість тисяч п\'ятсот сорок сім гривень тест'
    assert norm('просто 6 547 грн') == 'просто шість тисяч п\'ятсот сорок сім гривень.'
    assert norm('6 547 грн тест') == 'шість тисяч п\'ятсот сорок сім гривень тест'
    assert norm('Просто 6 547 грн тест') == 'Просто шість тисяч п\'ятсот сорок сім гривень тест'

    assert norm('просто 1 грв.') == 'просто одна гривня.'
    assert norm('1 грв. тест') == 'одна гривня тест'
    assert norm('Просто 1 грв. тест') == 'Просто одна гривня тест'
    assert norm('просто 1 грв.') == 'просто одна гривня.'
    assert norm('1 грв. тест') == 'одна гривня тест'
    assert norm('Просто 1 грв. тест') == 'Просто одна гривня тест'

    assert norm('просто 2 грв.') == 'просто дві гривні.'
    assert norm('2 грв. тест') == 'дві гривні тест'
    assert norm('Просто 2 грв. тест') == 'Просто дві гривні тест'
    assert norm('просто 2 грв.') == 'просто дві гривні.'
    assert norm('2 грв. тест') == 'дві гривні тест'
    assert norm('Просто 2 грв. тест') == 'Просто дві гривні тест'

    assert norm('просто 6 грв.') == 'просто шість гривень.'
    assert norm('6 грв. тест') == 'шість гривень тест'
    assert norm('Просто 6 грв. тест') == 'Просто шість гривень тест'
    assert norm('просто 6 грв.') == 'просто шість гривень.'
    assert norm('6 грв. тест') == 'шість гривень тест'
    assert norm('Просто 6 грв. тест') == 'Просто шість гривень тест'

    assert norm('просто 6 547 грв.') == 'просто шість тисяч п\'ятсот сорок сім гривень.'
    assert norm('6 547 грв. тест') == 'шість тисяч п\'ятсот сорок сім гривень тест'
    assert norm('Просто 6 547 грв. тест') == 'Просто шість тисяч п\'ятсот сорок сім гривень тест'
    assert norm('просто 6 547 грв.') == 'просто шість тисяч п\'ятсот сорок сім гривень.'
    assert norm('6 547 грв. тест') == 'шість тисяч п\'ятсот сорок сім гривень тест'
    assert norm('Просто 6 547 грв. тест') == 'Просто шість тисяч п\'ятсот сорок сім гривень тест'

    assert norm('просто 1 грв') == 'просто одна гривня.'
    assert norm('1 грв тест') == 'одна гривня тест'
    assert norm('Просто 1 грв тест') == 'Просто одна гривня тест'
    assert norm('просто 1 грв') == 'просто одна гривня.'
    assert norm('1 грв тест') == 'одна гривня тест'
    assert norm('Просто 1 грв тест') == 'Просто одна гривня тест'

    assert norm('просто 2 грв') == 'просто дві гривні.'
    assert norm('2 грв тест') == 'дві гривні тест'
    assert norm('Просто 2 грв тест') == 'Просто дві гривні тест'
    assert norm('просто 2 грв') == 'просто дві гривні.'
    assert norm('2 грв тест') == 'дві гривні тест'
    assert norm('Просто 2 грв тест') == 'Просто дві гривні тест'

    assert norm('просто 6 грв') == 'просто шість гривень.'
    assert norm('6 грв тест') == 'шість гривень тест'
    assert norm('Просто 6 грв тест') == 'Просто шість гривень тест'
    assert norm('просто 6 грв') == 'просто шість гривень.'
    assert norm('6 грв тест') == 'шість гривень тест'
    assert norm('Просто 6 грв тест') == 'Просто шість гривень тест'

    assert norm('просто 6 547 грв') == 'просто шість тисяч п\'ятсот сорок сім гривень.'
    assert norm('6 547 грв тест') == 'шість тисяч п\'ятсот сорок сім гривень тест'
    assert norm('Просто 6 547 грв тест') == 'Просто шість тисяч п\'ятсот сорок сім гривень тест'
    assert norm('просто 6 547 грв') == 'просто шість тисяч п\'ятсот сорок сім гривень.'
    assert norm('6 547 грв тест') == 'шість тисяч п\'ятсот сорок сім гривень тест'
    assert norm('Просто 6 547 грв тест') == 'Просто шість тисяч п\'ятсот сорок сім гривень тест'


def test_euro():
    # assert norm('просто 1€') == 'просто одне євро'
    # assert norm('1€ тест') == 'одне євро тест'
    # assert norm('Просто 1€ тест') == 'просто одне євро тест'
    # assert norm('просто 1€') == 'просто одне євро'
    # assert norm('1€ тест') == 'одне євро тест'
    # assert norm('Просто 1€ тест') == 'просто одне євро тест'

    assert norm('просто 2€') == 'просто два євро.'
    assert norm('2€ тест') == 'два євро тест'
    assert norm('Просто 2€ тест') == 'Просто два євро тест'
    assert norm('просто 2€') == 'просто два євро.'
    assert norm('2€ тест') == 'два євро тест'
    assert norm('Просто 2€ тест') == 'Просто два євро тест'

    assert norm('просто 6€') == 'просто шість євро.'
    assert norm('6€ тест') == 'шість євро тест'
    assert norm('Просто 6€ тест') == 'Просто шість євро тест'
    assert norm('просто 6€') == 'просто шість євро.'
    assert norm('6€ тест') == 'шість євро тест'
    assert norm('Просто 6€ тест') == 'Просто шість євро тест'

    assert norm('просто 6 547€') == 'просто шість тисяч п\'ятсот сорок сім євро.'
    assert norm('6 547€ тест') == 'шість тисяч п\'ятсот сорок сім євро тест'
    assert norm('Просто 6 547€ тест') == 'Просто шість тисяч п\'ятсот сорок сім євро тест'
    assert norm('просто 6 547€') == 'просто шість тисяч п\'ятсот сорок сім євро.'
    assert norm('6 547€ тест') == 'шість тисяч п\'ятсот сорок сім євро тест'
    assert norm('Просто 6 547€ тест') == 'Просто шість тисяч п\'ятсот сорок сім євро тест'


def test_unknown():
    assert norm('2-річний курс навчання.') == 'двох річний курс навчання.'
    assert norm('3-колісний велосипед.') == 'трьох колісний велосипед.'
    assert norm('5-стопний хорей.') == 'п\'яти стопний хорей.' 
    assert norm('2-годинна екскурсія.') == 'двох годинна екскурсія.'
    # assert norm('2-мільярдний прибуток.')
    # assert norm('5-мільйонний житель.')
    # assert norm('4-річна сестричка.')
    # assert norm('4-кілометровий пробіг.')
    # assert norm('4-тисячний тираж.')
    # assert norm('5-відсотковий кредит.')
    # assert norm('40-тонний вагон.')
    # assert norm('40-річний чоловік.')
    # assert norm('40-мяльярдний борг.')
    # assert norm('30-градусний мороз.')
    # assert norm('30-тисячний випуск.')
    # assert norm('12-річнийй син.')
    # assert norm('5-метрова глибина.') 
    # assert norm('3-значне число.')
    # assert norm('3-сотий журнал.')
    # assert norm('150-літній ювілей.')
    # assert norm('150-тисячна армія.') == 'сто пя\'тдесяти тисячна армія.'
    # assert norm('5-копійкова монета.')
    # assert norm('5-сотий кілометр.')

# def test_dash_replacments():
#     assert norm('Гаразд, — мовила Мак-Класкі.') == 'Гаразд, мовила Мак Класкі.'

def test_number_sentances():
    assert norm('1-го квітня, на 1-му поверсі Яринка загубила 2грн але знайшла 5€. Але її 4-річна сестричка забрала 50% її знахідки.') == \
    'першого квітня, на першому поверсі Яринка загубила дві гривні але знайшла п\'ять євро. Але її чотирьох річна сестричка забрала п\'ятдесят відсотків її знахідки.'

    assert norm('Також 2003 року щось там сталося і 40-річний чоловік помер. Його знайшли через 3 години.') == \
    'Також дві тисячі третього року щось там сталося і сорока річний чоловік помер. Його знайшли через три години.'

    assert norm('20% від 200 буде 40.') == \
    'двадцять відсотків від двісті буде 40.'

    assert norm('В будинуку №52.') == \
    'В будинуку номер п\'ятдесят два.'
 
    assert norm('Кожен, хто стоятиме у цій уявній вартівні у 2025-му, стоятиме в ній також і в 1914-му.') == \
    'Кожен, хто стоятиме у цій уявній вартівні у дві тисячі двадцять п\'ятому стоятиме в ній також і в одна тисяча дев\'ятсот чотирнадцятому.'

    # assert norm('Вони з\'являться аж десь біля 1926-го, а може й пізніше.') == 'Вони з\'являться аж десь біля одна тисяча дев\'ятсот двадцять шостого, а може й пізніше.'

    assert norm('Кроссліт 2025.') == 'Кроссліт дві тисячі двадцять п\'ять.'

    assert norm('200 тисяч британок опинилися 2002 року в лікарні через зловживання алкоголем. Кількість таких пацієнток 2010 року зросла понад удвічі.') == \
    'двісті тисяч британок опинилися дві тисячі другого року в лікарні через зловживання алкоголем. Кількість таких пацієнток дві тисячі десятого року зросла понад удвічі.'

    #assert norm('Розділ 8') == 'Розділ восьмий'
    #assert norm('Ми трималися там два дні і дві ночі, 130 чоловік з 16 льюісами.') == 'Ми трималися там два дні і дві ночі, сто тридцять чоловік з шістнадцятьма льюісами.'

    #assert norm('Кожного 2-го дня, я прокидався о 6 ранку.') == 'Кожного другого дня, я прокидався о шостій ранку.'


    


import os
import stable_whisper
from faster_whisper.tokenizer import Tokenizer
from narizaka.utils import convert_media


class FasterWhisperTranscriber():
    def __init__(self, device, device_index) -> None:

        self.model = stable_whisper.load_faster_whisper(
            'large-v2', device=device, device_index=device_index, cpu_threads=2)
        tokenizer = Tokenizer(
                    self.model.hf_tokenizer,
                    True,
                    task='transcribe',
                    language='uk',
                )
        self.number_tokens = [-1]
        self.number_tokens = self.number_tokens + [
            i 
            for i in range(tokenizer.eot)
            if all(c in "0123456789%№§IVXC$£₴€" for c in tokenizer.decode([i]).removeprefix(" "))
        ]

    def transcribe(self, audio_file):
        filename_16, _ = convert_media(audio_file, format='wav', sr=16000)
        result = self.model.transcribe_stable(filename_16, language='uk', regroup=True, verbose=None,
                                                        prepend_punctuations= "\"'“¿([{-«",
                                                        append_punctuations = "\"'.。,，!！?？:：”)]}、»", suppress_tokens=self.number_tokens)
        os.remove(filename_16)
        return result



#!/usr/bin/python3

import gi 
import os
import time
from queue import Queue, Empty

gi.require_version('Gst', '1.0')
from gi.repository import Gst, GLib

import logging
logger = logging.getLogger(__name__)
Gst.init(None)


class MediaInfo():
    def __init__(self):
        self.pipeline = Gst.Pipeline()
        self.src = Gst.ElementFactory.make("filesrc")
        self.pipeline.add(self.src)
        self.decbin = Gst.ElementFactory.make("decodebin")
        self.pipeline.add(self.decbin)
        self.conv = Gst.ElementFactory.make("audioconvert")
        self.pipeline.add(self.conv)
        self.res = Gst.ElementFactory.make("audioresample")
        self.pipeline.add(self.res)
        self.rg = Gst.ElementFactory.make("rganalysis")
        self.pipeline.add(self.rg)
        self.sink = Gst.ElementFactory.make("fakesink")
        self.pipeline.add(self.sink)

        self.src.link(self.decbin)
        self.conv.link(self.res)
        self.res.link(self.rg)
        self.rg.link(self.sink)
        self.decbin.connect("pad-added", self._on_pad_added)
        self.decbin.connect("pad-removed", self._on_pad_removed)

        self.rg.set_property("forced", True)
        #self.rg.set_property("reference-level", self.ref_lvl)

        bus = self.pipeline.get_bus()
        bus.add_signal_watch()
        bus.connect("message", self._on_message)

    def _on_pad_added(self, decbin, new_pad):
        sinkpad = self.conv.get_compatible_pad(new_pad, None)
        if sinkpad is not None:
            new_pad.link(sinkpad)

    def _on_pad_removed(self, decbin, old_pad):
        peer = old_pad.get_peer()
        if peer is not None:
            old_pad.unlink(peer)

    def _on_message(self, bus, msg):
        if msg.type == Gst.MessageType.TAG:
            tags = msg.parse_tag()
            self.tl = Gst.TagList.new_empty()
            def handle_tag(taglist, tag, userdata):
                if tag == Gst.TAG_TRACK_GAIN:
                    self.tl.add_value(Gst.TagMergeMode.APPEND, Gst.TAG_TRACK_GAIN,
                                      taglist.get_double(tag)[1])
                elif tag == Gst.TAG_TRACK_PEAK:
                   self.tl.add_value(Gst.TagMergeMode.APPEND, Gst.TAG_TRACK_PEAK,
                                     taglist.get_double(tag)[1])
                elif tag == Gst.TAG_REFERENCE_LEVEL:
                    self.tl.add_value(Gst.TagMergeMode.APPEND, Gst.TAG_REFERENCE_LEVEL,
                                      taglist.get_double(tag)[1])
            tags.foreach(handle_tag, None)
        elif msg.type == Gst.MessageType.EOS:
            self.pipeline.set_state(Gst.State.NULL)
            self.mainloop.quit()

    def get(self, fname):
        self.src.set_property("location", fname)
        self.pipeline.set_state(Gst.State.PLAYING)
        self.mainloop = GLib.MainLoop()
        self.mainloop.run()
        return self.tl


class Segmenter():
    def __init__(self, sr):

        self.sr = sr
        self.info = MediaInfo()

        self.splits = Queue()
        self.current_segment = None
        self.index = 0
        self.prerolled = False

        
        self.pipeline = Gst.Pipeline()
        self.mainloop = GLib.MainLoop()
        self.bus = self.pipeline.get_bus()
        self.pipeline.bus.add_signal_watch()
        self.bus.connect('message', self.onMessage, self.mainloop)
        
        self.src = Gst.ElementFactory.make("filesrc")
        self.pipeline.add(self.src)
        
        self.decoder = Gst.ElementFactory.make("decodebin")
        self.pipeline.add(self.decoder)
        self.decoder.set_property('caps', Gst.Caps.from_string('audio/x-raw'))
        self.decoder.connect("pad-added", self.on_pad_added)
        
        self.src.link(self.decoder)

        self.aconv = Gst.ElementFactory.make('audioconvert')
        self.pipeline.add(self.aconv)
        self.decoder.link(self.aconv)

        self.dein = Gst.ElementFactory.make('deinterleave')
        self.pipeline.add(self.dein)
        self.deint_conn = self.dein.connect("pad-added", self.on_channel_added)
        self.dein.connect("no-more-pads", self.no_more_pads)
        self.aconv.link(self.dein)

        self.encoder = Gst.ElementFactory.make("wavenc")
        self.pipeline.add(self.encoder)

        self.anorm = Gst.ElementFactory.make("rgvolume")
        self.pipeline.add(self.anorm)
        self.anorm.set_property("album-mode", False)
        #self.anorm.set_property("pre-amp", 12)

        self.capsfilter = Gst.ElementFactory.make('capsfilter')
        self.pipeline.add(self.capsfilter)
      

        self.resample = Gst.ElementFactory.make('audioresample')
        self.pipeline.add(self.resample)
        self.resample.set_property('quality', 10)

        self.capsfilter.set_property('caps', Gst.Caps.from_string(f'audio/x-raw,rate={sr}'))
        self.resample.link(self.capsfilter)
        
        self.resample.link(self.capsfilter)
        self.capsfilter.link(self.anorm)


        self.anorm.link(self.encoder)
        

        
        self.filesink = Gst.ElementFactory.make("filesink")
        self.filesink.set_property('buffer-mode', 2)
        self.pipeline.add(self.filesink)
        self.encoder.link(self.filesink)


        src_filesink_pad = self.filesink.get_static_pad('sink')
        src_filesink_pad.add_probe(Gst.PadProbeType.EVENT_FLUSH, self.pad_filesink_probe)
        

        self.src.set_state(Gst.State.NULL)


    
    def on_pad_added(self, el, pad):
        if self.prerolled:
            return
        sink = self.aconv.get_static_pad('sink')
        pad.link(sink)
    

    def no_more_pads(self, el):
        if self.prerolled:
            return
        self.custom_message('do_seek')

    def custom_message(self, name):
        custom_structure = Gst.Structure.new_empty(name)
        custom_message = Gst.Message.new_application(None, custom_structure)
        self.bus.post(custom_message)

    def on_channel_added(self, el, pad):
        if self.prerolled:
            return
        pad.link(self.resample.get_static_pad('sink'))

    
    def run(self, location, output_folder):
        if not self.splits.empty():
            self.rp_event = Gst.Event.new_tag(self.info.get(location))
            self.pipeline.set_state(Gst.State.NULL)
            self.output_folder = output_folder
            os.makedirs(output_folder, exist_ok=True)
            self.src.set_property("location", location)
            
            self.current_segment = self.splits.get_nowait()
            self.filesink.set_property("location", os.path.join(self.output_folder, self.construct_name(self.current_segment[2])))
            self.prerolled = False
            self.pipeline.set_state(Gst.State.PAUSED)

            self.mainloop.run()
            

    def construct_name(self, index):
        return f'segment_{index:06}.wav'
    
    def save(self, start_time, end_time):
        self.splits.put_nowait((start_time, end_time, self.index))
        file_path = self.construct_name(self.index)
        self.index += 1
        return file_path

    def pad_filesink_probe(self, pad, info):
            event = info.get_event()
            if event.type == Gst.EventType.FLUSH_STOP:
                self.filesink.set_state(Gst.State.NULL)
                self.filesink.set_property("location", os.path.join(self.output_folder, self.construct_name(self.current_segment[2])))
                self.filesink.set_state(Gst.State.PLAYING)
                
            return Gst.PadProbeReturn.OK
    
    def do_seek(self):
        seek_flags = Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE |  Gst.SeekFlags.SEGMENT
        res = self.pipeline.seek(1.0, Gst.Format.TIME, seek_flags, Gst.SeekType.SET, self.current_segment[0] * Gst.SECOND, Gst.SeekType.SET, self.current_segment[1] * Gst.SECOND)
        if not res:
            raise Exception('Can\'t seek')
 
    
    def onMessage(self, bus: Gst.Bus, message: Gst.Message, loop: GLib.MainLoop):
        
        mtype = message.type
        if mtype == Gst.MessageType.EOS:
            self.pipeline.set_state(Gst.State.NULL)
            self.mainloop.quit()
        elif message.type == Gst.MessageType.APPLICATION:
            if message.get_structure().get_name() == "do_seek":
                self.do_seek()
                self.prerolled = True
                self.pipeline.set_state(Gst.State.PLAYING)

        elif mtype == Gst.MessageType.SEGMENT_DONE:

            try:
                self.current_segment = self.splits.get_nowait()
                self.custom_message('do_seek')
                self.encoder.set_state(Gst.State.NULL)
                self.encoder.set_state(Gst.State.PLAYING)
            except Empty:
                self.bus.post(Gst.Message.new_eos())
        elif mtype == Gst.MessageType.STREAM_START:
            self.resample.get_static_pad('src').push_event(self.rp_event)

        elif mtype == Gst.MessageType.ERROR:
            err, debug = message.parse_error()
            print("Error: %s: %s\n" % (err, debug))
            bus.post(Gst.Message.new_eos())

#Example
# s = Segmenter(sr=24000)
# s.save(0.05, 18.05)
# s.save(18.85, 21.450000000000003)
# s.save(21.8, 28.750000000000004)
# s.run('test_data/ggg.mp3', 'testoutput')
# s.save(23.8, 28.750000000000004)
# s.run('test_data/ggg.mp3', 'testoutput')




import regex
import unicodedata
import os
import pathlib
import json


from narizaka.textbook import TextBook
from narizaka.splitter import Splitter
from narizaka import utils
from narizaka.languages.uk.textnormalizer import norm
from fuzzysearch import find_near_matches
from csv import DictWriter, QUOTE_MINIMAL
from faster_whisper.utils import format_timestamp
from narizaka.segmenter import Segmenter
import torchaudio
from torchaudio.functional import resample
from stable_whisper.result import WordTiming


bad_text = regex.compile('[0-9\p{L}--[а-яіїєґ]]', regex.VERSION1|regex.IGNORECASE)

class Aligner():
    def __init__(self, args) -> None:
        self.output = args.o
        self.sr = args.sr
        self.splitter = Splitter()
        self.columns = args.columns.split(',')

    def pases_filter(self, segment):
        if segment['end']-segment['start'] < 1.0 or segment['end']-segment['start'] > 35:
            #print('Skipped because of length')
            return False
        if bad_text.search(segment['sentence']):
            #print('Skipped because contains inapropirate characters')
            return False
        return True

    
    def _base_norm(self, text):
        return norm(text).split(' ')
    
    def _norm_word(self, text):
        text = regex.sub(r'[^\p{L}\p{N}]', '', text)
        return text.lower()
    
    def _norm_for_corpus(self, text):
        text = regex.sub(r'[^\p{L}\p{N}\?\!\,\.\-\: ]', '', text)
        return text
    
    def normalize(self, text):
        words = self._base_norm(text)
        normalized = ''
        for word in words:
            nword = self._norm_word(word)
            if nword:
                normalized += ' ' + nword
        return normalized

    def _feed(self, text) -> None:
        words = self._base_norm(text)
        for word in words:
            nword = self._norm_word(word)
            if nword:
                self.denorm.append(word)
                word_pos = len(self.norm_text.split(' '))-1
                chars_in_word = len(nword)
                for _ in range(chars_in_word+1):
                    self.normpos.append(word_pos)
                self.norm_text += ' ' + nword
            elif self.denorm:
                self.denorm[-1] += ' '+word

    def get_denorm(self, pos, norm_text):
        nwords = len(norm_text.split(' '))-1
        start_pos = self.current_pos+pos
        denorm_words_list = self.denorm[self.normpos[start_pos]:self.normpos[start_pos]+nwords]
        return ' '.join(denorm_words_list)
            
    def current_norm_text(self):
        if len(self.norm_text) < self.current_pos + 20000:
            text = self.book.more_text()
            self._feed(text)
        return self.norm_text[self.current_pos:]
        
    def find_match(self, text):
        norm_text = self.normalize(text)
        num_words = len(norm_text.split(' '))
    
        match = {
            'asr': norm_text,
            'matched': False,
        }
        if not norm_text:
            match['book_text'] = self.current_norm_text()[:20 ] + "..."
            return match
        matches = find_near_matches(norm_text, self.current_norm_text(), max_l_dist=int(num_words*0.2))
        if matches and matches[0].matched:
            matched = min(matches, key=lambda m: m.start)
            
            match['sentence'] = self._norm_for_corpus(self.get_denorm(matched.start, matched.matched))
            match['sentence_norm'] = matched.matched
            match['start'] = matched.start
            match['end'] = matched.end
            match['distance'] = matched.dist
            match['matched'] = True
            if num_words > 4 or matched.start < 20:
                self.current_pos += matched.end
        else:
            book_text = self.current_norm_text()[:200 ] + "..."
            match['book_text'] = book_text
        return match


    def run(self, pair):
        self.normpos = []
        self.denorm = []
        self.norm_text = ''
        self.current_pos = 0
        self.book = TextBook(pair.text_book_path)

        self.recognised_duration = 0.0

        audio_output = self.output / self.book.name
        if not audio_output.exists():
            os.makedirs(audio_output, exist_ok=True)
    
            
        dataset_file = self.output.joinpath(self.book.name+'.txt')
        log_file = open(self.output.joinpath(self.book.name+'.log'), 'w')
        log_file.write(f'\nStarting book {self.book.name}\n')
        dfp = dataset_file.open("w")
        ds = DictWriter(
            dfp,
            fieldnames=self.columns,
            extrasaction='ignore',
            quoting=QUOTE_MINIMAL,
            delimiter='|'
        )

        segmenter = Segmenter(sr=self.sr)
        for audio_file, words in pair.audio_book.transcription():
            log_file.write(f'Starting audiofile: {audio_file}\n')
            file_16, _ = utils.convert_media(audio_file, format='wav', sr=16000)
            segments = self.splitter.split_to_segments(file_16, words)
            for segment in segments:
                if not segment:
                    continue
                log_file.write(f'\n{format_timestamp(segment["start"])} -> {format_timestamp(segment["end"])}: {segment["text"]}\n')

                match = self.find_match(segment["text"])
                if match.get('matched'):
                    segment['sentence'] = match["sentence"]
                    log_file.write(f'MATCHED: {match["sentence"]}\n')
                    if not self.pases_filter(segment):
                        continue

                    start = float(segment['start'])
                    end = float(segment['end'])
                    filename = segmenter.save(start_time=start, end_time=end)
                    
                    segment['duration'] = segment['end'] - segment['start']
                    segment['audio'] = os.path.join(self.book.name, filename)
                    segment['speaker_id'] = pair.audio_book.speaker_id or '0'

                    ds.writerow(segment)
                    self.recognised_duration += segment['duration']

                else:
                    log_file.write(f'NOT MATCHED: {match["book_text"]}\n')
            os.remove(file_16)
            segmenter.run(str(audio_file), output_folder=audio_output)
        dfp.close()
        log_file.close()
        return self.recognised_duration, pair.audio_book.duration, self.book.name



import os
import tempfile
import regex
import magic
import pathlib

import xml.etree.ElementTree as ET

class TextBook:
    def __init__(self, path, min_text_length=20000) -> None:
        super().__init__()
        self.filename = path
        self.name = path.stem
        self.min_text_length = min_text_length
        self.temp_file = None
        self.can_skip = False
        if magic.from_file(filename=path, mime=True) == 'application/epub+zip':
            self.can_skip = True
        
        if self._is_fb2(path):
            self.path = path
        else:
            fl, self.temp_file = tempfile.mkstemp(suffix='.fb2')
            os.close(fl)
            os.system(f'pandoc "{path}" -o {self.temp_file}')
            self.path = self.temp_file
        self.iter = ET.parse(self.path).getroot().find('{http://www.gribuser.ru/xml/fictionbook/2.0}body').iter()

    
    def _is_fb2(self, filename: pathlib.Path)-> bool:
        mimetype = magic.from_file(filename=filename, mime=True)
        return mimetype == 'text/xml'
    
    def norm(self, text):
        text = regex.sub(r'[᠆‐‑‒–—―⁻₋−⸺⸻]', '-', text)
        text = regex.sub(r'^\s*?\-', '', text)
        text = regex.sub(r'\[.*?\]', '', text)
        text = regex.sub(r'\s+\.', '. ', text)
        return text
    
    def _get_text(self, el):
        tag = el.tag
        if not isinstance(tag, str) and tag is not None:
            return
        text = ''
        if el.text:
            text += el.text
        
        for e in el:
            if not e.tag.endswith('}a'):
                for s in self._get_text(e):
                    text += s
            if e.tail:
               text += e.tail
        return text

    
    def more_text(self):
        text = ''
        skip = False
        for i in self.iter:
            if i.tag.endswith('}p') and not skip:
                text += self.norm(self._get_text(i)) + ' '
                if len(text) >= self.min_text_length:
                    break
            elif i.tag.endswith('}empty-line') and self.can_skip:
                skip = True
            elif i.tag.endswith('}p') and skip and i.text == None and i.tail == None and not "".join(i.itertext()):
                skip = False
        return text
    
    def __del__(self):
        if self.temp_file:
            os.remove(self.temp_file)



import sys
import pytest
from narizaka import narizaka


def test_main():
    sys.argv = ['narizaka', '-h']
    with pytest.raises(SystemExit) as excinfo:
        narizaka.run()
    assert excinfo.value.code == 0